/*
Package goth provides a simple, clean, and idiomatic way to write authentication
packages for Go web applications.

This package was inspired by https://github.com/intridea/omniauth.

See the examples folder for a working application that lets users authenticate
through Twitter or Facebook.
*/
package goth
